package qnd;

import java.io.BufferedWriter;
import java.io.File;
import java.util.ArrayList;
import java.util.Hashtable;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.SAXParserAdapter;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

public class ConvertIntermediateXML extends SAXParserAdapter{
    int margin = 1;
    String filename;
    String tagName;
    ArrayList<String> conj = new ArrayList<String>();
    ArrayList<String> part = new ArrayList<String>();
    boolean convert;
    boolean cases;
    BufferedWriter bw;
    String ruleNumber;
    String sutra;
    boolean ignore = false;
    boolean both = false;
    String currentTag;
    
    public ConvertIntermediateXML (String filename, String tag) {
        this.filename = filename;
        this.tagName = tag;
    }
    
    public void parse () throws Exception {
        bw = IOAssistance.openWriter("result.xml");
        bw.write("<rules>\n");
        SAXParserFactory factory = SAXParserFactory.newInstance();
        SAXParser saxParser = factory.newSAXParser();

        // See http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description
        // Uncomment the following to turn on comment processing.
        // http://docs.oracle.com/javaee/1.4/tutorial/doc/JAXPSAX11.html
        // saxParser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
        try {
            saxParser.parse(new File(filename), this);
        } catch (Exception ex) {
            System.out.format("%d. %s: %s\n", locator.getLineNumber(), filename, ex.toString());
        }        
        bw.write("</rules>");
        bw.close();
    }
    
    @Override
    public void startDocument () throws SAXException {
    }

    @Override
    public void endDocument () throws SAXException {
        try {
        } catch(Exception ex) {
            throw new SAXException(ex);
        }
        /*
        System.out.println("Conjunction\n");
        for (String s : conj) {
            System.out.format("\t%s\n", s);
        }

        System.out.println("Participle\n");
        for (String s : part) {
            System.out.format("\t%s\n", s);
        }
        */
    }
    
    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        push(qName, atts);
        if (ignore) {
            return;
        }
        try {
            if (qName.equals("rule")) {
                margin = 1;
                both = false;
                if (atts.getValue("completed") != null) {
                    convert = false;
                } else if (atts.getValue("both") != null) {
                    both = true;
                    convert = true;
                } else {
                    convert = true;
                }
                if (!convert) {
                    ignore = true;
                    return;
                }
                
                cases = atts.getValue("cases") != null;

                if (convert) {
                    ruleNumber = atts.getValue("n");
                    sutra = atts.getValue("s");
                    bw.write(String.format(indent("<rule n=\"%s\" s=\"%s\">\n", margin), ruleNumber, sutra));
                    if (cases) {
                        margin++;
                        bw.write(String.format(indent("<cases>\n", margin)));
                        margin++;
                    }
                }
                
            } else if (qName.equals("description")) {
                String tag = atts.getValue("tag");

                if (both && !tag.equals("par")) {
                    return;
                }
                
                if (cases) {
                    bw.write(indent("<case>\n", margin));
                }
                ArrayList<String> pUrva = new ArrayList<String>();
                ArrayList<String> sTAnin = new ArrayList<String>();
                ArrayList<String> para = new ArrayList<String>();
                ArrayList<String> AdeSa = new ArrayList<String>();
                ArrayList<String> errors = new ArrayList<String>();
                String ruleset = atts.getValue("ruleset");
                String source = atts.getValue("source");
                String target = atts.getValue("target");
                String lcontext = atts.getValue("lcontext");
                String rcontext = atts.getValue("rcontext");
                String optional = atts.getValue("optional");
                String root = atts.getValue("root");
                String affix = atts.getValue("affix");
                boolean rootPlaced = false;
                boolean affixPlaced = false;
                if (source != null) {
                    sTAnin.add(String.format("phone=\"%s\"", source));
                }
                if (target != null) {
                    AdeSa.add(String.format("phone=\"%s\"", target));
                }
                
                if (lcontext != null) {
                    pUrva.add(String.format("phone=\"%s\"", lcontext));
                }
                
                if (root != null) {
                    if (tag.equals("par")) {
                        if (ruleset.equals("endings_changes") || ruleset.equals("affix_modifications")) {
                            pUrva.add(String.format("saYjYA=\"DAtu[slp=%s]\"", root));
                            rootPlaced = true;
                        }
                    } else {
                        if (ruleset.equals("endings_changes")) {
                            pUrva.add(String.format("saYjYA=\"DAtu[slp=%s]\"", root));
                            rootPlaced = true;
                        }
                    }
                    if (!rootPlaced) {      // see if root goes in sTAnin
                        if (tag.equals("par")) {
                            if (ruleset.equals("root_replacements")
                                    || ruleset.equals("root_modifications")
                                    || ruleset.equals("reduplication")
                                    || ruleset.equals("stem_changes")
                                    || ruleset.equals("consonant-stem_changes")) {
                                sTAnin.add(String.format("saYjYA=\"DAtu[slp=%s\"", root));
                                rootPlaced = true;
                            }
                        } else {
                            if (ruleset.equals("root_replacements")
                                    || ruleset.equals("stem_changes")
                                    || ruleset.equals("consonant-stem_changes")) {
                                sTAnin.add(String.format("saYjYA=\"DAtu[slp=%s\"", root));
                                rootPlaced = true;
                            }
                        }
                    }
                }
                
                
                if (affix != null) {
                    if (tag.equals("par")) {
                        if (ruleset.equals("endings_changes") || ruleset.equals("affix_modifications")) {
                            sTAnin.add(String.format("saYjYA=\"pratyaya[slp=%s]\"", affix));
                            affixPlaced = true;
                        }
                    } else {
                        if (ruleset.equals("endings_changes")) {
                            sTAnin.add(String.format("saYjYA=\"pratyaya[slp=%s]\"", affix));
                            affixPlaced = true;
                        }
                    }
                    if (!affixPlaced) {
                        if (tag.equals("par")) {
                            if (ruleset.equals("root_replacements") ||
                                ruleset.equals("root_modifications") ||
                                ruleset.equals("reduplication") ||
                                ruleset.equals("stem_changes") ||
                                ruleset.equals("consonant-stem_changes")) {
                                para.add(String.format("saYjYA=\"pratyaya[slp=%s]\"", affix));
                                affixPlaced = true;
                            }
                        } else {
                            if (ruleset.equals("root_replacements") ||
                                ruleset.equals("stem_changes") ||
                                ruleset.equals("consonant-stem_changes")) {
                                para.add(String.format("saYjYA=\"pratyaya[slp=%s]\"", affix));
                                affixPlaced = true;
                            }
                        }
                    }
                }
                
                if (!rootPlaced && root != null) {
                    errors.add("Where to put root?");
                }
                
                if (!affixPlaced && affix != null) {
                    errors.add("Where to put affix?");
                }
                
                if (pUrva.size() > 0) {
                    bw.write(indent("<pUrva", margin+1));
                    for (String s : pUrva) {
                        bw.write(String.format(" %s", fixup(s)));
                    }
                    bw.write("/>\n");
                }
                
                if (sTAnin.size() > 0) {
                    bw.write(indent("<sTAnin", margin+1));
                    for (String s : sTAnin) {
                        bw.write(String.format(" %s", fixup(s)));
                    }
                    bw.write("/>\n");
                }

                if (para.size() > 0) {
                    bw.write(indent("<para", margin+1));
                    for (String s : para) {
                        bw.write(String.format(" %s", fixup(s)));
                    }
                    bw.write("/>\n");
                }
                
                if (AdeSa.size() > 0) {
                    bw.write(indent("<AdeSa", margin+1));
                    for (String s: AdeSa) {
                        bw.write(String.format(" %s", fixup(s)));
                    }
                    bw.write("/>\n");
                }
                
                if (errors.size() > 0) {
                    for (String s: errors) {
                        bw.write(String.format(indent("<error>%s</error>\n", margin+1), s));
                    }
                    
                }

                if (cases) {
                    bw.write(indent("</case>\n", margin));
                }
            }
        } catch (Exception ex) {
        }
    }
    
    
    private String indent (String s, int margin) {
        return "      ".substring(0, margin) + s;
    }

    private String fixup (String s) {
        return s.replaceAll("<", "&lt;");
    }
    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
            try {
                if (qName.equals("rule")) {
                    if (!ignore) {
                        if (cases) {
                            bw.write(indent("</cases>\n", margin-1));
                        }
                        margin = 1;
                        bw.write(indent("</rule>\n\n", margin));
                    }
                    ignore = false;
                }
            } catch (Exception ex) {
            }
        pop();
    }

    @Override
    public void characters(char[] ch, int start, int length) throws SAXException {
        String path = buildPath();
        StringBuilder buffer = buffers.get(path);
        if (buffer == null) {
            buffer = new StringBuilder();
            buffers.put(path, buffer);
        }
        buffer.append(new String(ch, start, length));
    }
}
