package qnd;

import java.io.File;
import java.util.ArrayList;
import java.util.Hashtable;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.SAXParserAdapter;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

public class ConvertXMLRuleFormat extends SAXParserAdapter{
    public static Hashtable<String, ArrayList<RuleInfo>> rules = new Hashtable<String, ArrayList<RuleInfo>>();
    String filename;
    String tag;
    String currentRuleset;
    Regex reNumber = new Regex("", "^(\\d\\.\\d\\.\\d+) ?(.*)");
    
    public ConvertXMLRuleFormat(String filename, String tag) {
        this.filename = filename;
        this.tag = tag;
    }
    
    public void parse () throws Exception {
        SAXParserFactory factory = SAXParserFactory.newInstance();
        SAXParser saxParser = factory.newSAXParser();

        // See http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description
        // Uncomment the following to turn on comment processing.
        // http://docs.oracle.com/javaee/1.4/tutorial/doc/JAXPSAX11.html
        // saxParser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
        try {
            saxParser.parse(new File(filename), this);
        } catch (Exception ex) {
            System.out.format("%d. %s: %s\n", locator.getLineNumber(), filename, ex.toString());
        }        
    }
    
    @Override
    public void startDocument () throws SAXException {
    }

    @Override
    public void endDocument () throws SAXException {
    }
    
    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        push(qName, atts);
        if (qName.equals("ruleset")) {
            currentRuleset = atts.getValue("name");
        } else if (qName.equals("rule")) {
            if (currentRuleset != null) {
                String c = atts.getValue("c");

                if (c != null) {
                    reNumber.reset(c);
                    if (reNumber.find()) {
                        String ruleNumber = reNumber.getGroup(1);
                        String sutra = reNumber.getGroup(2);
                        String source = atts.getValue("source");
                        String target = atts.getValue("target");
                        String lcontext = atts.getValue("lcontext");
                        String rcontext = atts.getValue("rcontext");
                        String root = atts.getValue("root");
                        String affix = atts.getValue("affix");
                        String optional = atts.getValue("optional");
                        
                        ArrayList<RuleInfo> a = rules.get(ruleNumber);
                        if (a == null) {
                            a = new ArrayList<RuleInfo>();
                            rules.put(ruleNumber, a);
                        }
                        a.add(new RuleInfo(sutra, tag, currentRuleset, source, target, lcontext, rcontext, root, affix, optional));
                    }
                }
            }
        }
    }
    
    private void add (StringBuilder sb, String value, String name) {
        if (value == null) {
            value = "null";
        }
        sb.append(String.format("%s=%s ", name, value));
    }
    
    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
        pop();
    }
    
    @Override
    public void characters(char[] ch, int start, int length) throws SAXException {
        String path = buildPath();
        StringBuilder buffer = buffers.get(path);
        if (buffer == null) {
            buffer = new StringBuilder();
            buffers.put(path, buffer);
        }
        buffer.append(new String(ch, start, length));
    }
}
