package qnd;

import java.io.BufferedWriter;
import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Hashtable;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.SAXParserAdapter;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

public class DependencyTags extends SAXParserAdapter {
    private static StringBuilder html = new StringBuilder();
    private static Hashtable<String, Integer> h = new Hashtable<String, Integer>();
    private final static Hashtable<String, String[]> hv = new Hashtable<String, String[]>();
    static {
        html.append("<html><head><title>Ramopakhyana</title><head><body>");
        String ramo = "/Users/ralph/SanskritLibrary/data/texts/rAmopAKyAna.slp1";
        try {
            String text = IOAssistance.readEntireFile(ramo);
            String[] verses = text.split("\r");
            for (String verse : verses) {
                String[] padas = verse.split("\\s+");
                hv.put(padas[0], padas);
            }
        } catch (Exception ex) {
            System.out.println(ex.toString());
        }
    }
    private BufferedWriter out;
    private String[] padas;
    private ArrayList<String> relations = new ArrayList<String>();
    private ArrayList<Integer> refs = new ArrayList<Integer>();
    private String versenum;
    
    public DependencyTags (String filename, String versenum, BufferedWriter out)  throws Exception {
        this.out = out;
        this.versenum = versenum;
        padas = hv.get(versenum);
        if (padas == null) {
            System.out.format("%s not found\n", versenum);
        }
        SAXParserFactory factory = SAXParserFactory.newInstance();
        // Parse the input
        SAXParser saxParser = factory.newSAXParser();
        // See http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description
        // Uncomment the following to turn on comment processing.
        // http://docs.oracle.com/javaee/1.4/tutorial/doc/JAXPSAX11.html
        // saxParser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
        try {
            saxParser.parse( new File(filename), this);
        } catch (Exception ex) {
            throw ex;
        }        
    }
    
    public static ArrayList<String> getTags () {
        ArrayList<String> a = new ArrayList<String>();
        
        Enumeration<String> en = h.keys();
        while (en.hasMoreElements()) {
            a.add(en.nextElement());
        }
        Collections.sort(a);
        return a;
    }
    
    public static String getHTML () {
        return html.toString();
    }
    
    private void write(String s) throws Exception {
        if (out != null) {
            out.write(s);
        }
    }
    
    @Override
    public void endDocument () {
        if (padas.length-1 != relations.size()) {
            System.out.println(versenum);
        } else {
            int len = padas.length;
            html.append("<p>");
            for (int i=0; i<len; i++) {
                html.append(padas[i]);
                html.append(" ");
            }
            html.append("</p>\n");
            
            len--;
            for (int i=0; i<len; i++) {
                int n = refs.get(i);
                if (n == 0) {
                    html.append("<span style='color:red'>");
                }
                html.append(padas[i+1]);
                html.append(" --");
                html.append(relations.get(i));
                html.append("--> ");
                if (n != 0) {
                    html.append(padas[n]);      // note: n is 1-based
                }
                if (n==0) {
                    html.append("</span>");
                }
                html.append("<br/>\n");
            }
        }
        html.append("<br/>");
    }
    
    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        try {
            push(qName, atts);
            if (qName.equals("links")) {
                write(String.format("<links>\n"));
            } else if (qName.equals("relations")) {
                write(String.format("\t<relations>\n"));
            } else if (qName.equals("refs")) {
                write(String.format("\t<refs>\n"));
            }
        } catch (Exception ex) {
            throw new SAXException(ex);
        }
    }
    
    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
        try {
            if (qName.equals("links")) {
                write(String.format("</links>\n"));
            } else if (qName.equals("relations")) {
                write(String.format("\t</relations>\n"));
            } else if (qName.equals("refs")) {
                write(String.format("\t</refs>\n"));
            } else {
                String path = buildPath();
                StringBuilder buffer = buffers.get(path);
                String value = null;
                if (buffer != null) {
                    value = buffer.toString();
                    buffer.setLength(0);
                } else {
                    value = "";
                }

                if (qName.startsWith("r")) {
                    Integer count = h.get(value);
                    if (count == null) {
                        count = new Integer(0);
                    }
                    h.put(value, new Integer(count.intValue() + 1));
                }
                write(String.format("\t\t<%s>%s</%s>\n", qName, value, qName));
                
                if (qName.startsWith("r")) {
                    relations.add(value);
                } else if (qName.startsWith("t")) {
                    try {
                        refs.add(Integer.parseInt(value));
                    } catch (Exception ex) {
                        refs.add(0);
                    }
                }
            }
            pop();
        } catch (Exception ex) {
            throw new SAXException(ex);
        }
    }

    
    @Override
    public void characters (char[] ch, int start, int end) {
        String path = buildPath();
        StringBuilder buffer = buffers.get(path);
        if (buffer == null) {
            buffer = new StringBuilder();
            buffers.put(path, buffer);
        }
        buffer.append(new String(ch, start, end));
    }

}
