package qnd;

import java.io.File;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.sanskritlibrary.SAXParserAdapter;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

public class DependencyTemplate extends SAXParserAdapter {
    private StringBuilder sb = new StringBuilder ();
    private StringBuilder comment = new StringBuilder("<!--");
    private boolean appendForm = true;
    private String currentRecord = null;
    private String nextRecord = null;
    
    public DependencyTemplate (String filename)  throws Exception {
        SAXParserFactory factory = SAXParserFactory.newInstance();

        // Parse the input
        SAXParser saxParser = factory.newSAXParser();
        // See http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description
        // Uncomment the following to turn on comment processing.
        // http://docs.oracle.com/javaee/1.4/tutorial/doc/JAXPSAX11.html
        // saxParser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
        try {
            saxParser.parse( new File(filename), this);
        } catch (Exception ex) {
            throw ex;
        }        
    }
    
    public String getDocument () {
        return sb.toString();
    }
    
    @Override
    public void endDocument () {
    }
    
    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        push(qName, atts);
        if (qName.equals("body")) {
            sb.append("<?xml version='1.0' encoding='UTF-8'?>\n");
            sb.append("<body>\n");        
        }
    }
    
    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
        String path = buildPath();
        StringBuilder buffer = buffers.get(path);
        String value = null;
        if (buffer != null) {
            value = buffer.toString();
            buffer.setLength(0);
        } else {
            value = "";
        }
        if (qName.equals("body")) {
            generateRec();
            sb.append("</body>\n");
        } else if (qName.equals("bcvnum")) {   
            if (currentRecord != null && !currentRecord.equals(value)) {
                generateRec();
            }
            nextRecord = currentRecord;
            currentRecord = value;
        } else if (qName.equals("pnum")) {
            if (!value.contains(".")) {
                comment.append(value);
                comment.append(". ");
                appendForm = true;
            }
        } else if (qName.equals("form")) {
            if (appendForm) {
                appendForm = false;
                comment.append(value);
                comment.append(" ");
            }
        }
        pop();
    }

    private void generateRec () {
                sb.append("<rec>\n");
                comment.append("-->");
                sb.append(comment);
                comment.setLength(0);
                comment.append("<!--");
                sb.append(String.format("\n\t<bcvnum>%s</bcvnum>\n", nextRecord));
                sb.append("\t<virtualNode id='' type=''>\n");
                sb.append("\t\t<node id=''/>\n");
                sb.append("\t</virtualNode>\n");
                sb.append("\t<arrows>\n");
                for (int i=0; i<5; i++) {
                    sb.append("\t\t<arrow source='' target='' type=''/>\n");
                }
                sb.append("\t</arrows>\n");            
                sb.append("</rec>\n\n");        
    }
    
    @Override
    public void characters (char[] ch, int start, int end) {
        String path = buildPath();
        StringBuilder buffer = buffers.get(path);
        if (buffer == null) {
            buffer = new StringBuilder();
            buffers.put(path, buffer);
        }
        buffer.append(new String(ch, start, end));
    }

}
