/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package qnd;

// This does nothing. It is a template for writing a class that parses XML

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Hashtable;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.SAXParserHandler;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
//DefaultHandler implements ContentHandler, DTDHandler, EntityResolver, ErrorHandler
//import org.xml.sax.helpers.DefaultHandler;

//DefaultHandler2 implements ContentHandler, DTDHandler, EntityResolver, ErrorHandler, DeclHandler, EntityResolver2, LexicalHandler
//http://docs.oracle.com/javase/1.5.0/docs/api/org/xml/sax/ext/DefaultHandler2.html
import org.xml.sax.ext.DefaultHandler2;         // this allows processing of comments because it implements LexicalHandler


// SAXParserHandler is an interface that forces us to include required methods.
// Serializable is optional. Have made local variables that don't need to be serialized transient.
public class FixupOCRXML extends DefaultHandler2 implements SAXParserHandler, Serializable {
    transient Locator locator;
    transient ArrayList<String> pathStack = new ArrayList<String>();
    transient Hashtable<String, StringBuilder> buffers = new Hashtable<String, StringBuilder>();
    transient Regex re = new Regex("", "/(.*)\\.jpg$");
    private String inputXMLFile;
    private String outputXMLFile;
    private String currentJPG;
    private LineInfo currentLineInfo;
    private int currentCropNumber = 0;
    StringBuilder badLines = new StringBuilder();
    ManuscriptInfo currentManuscript;
    ArrayList<ManuscriptInfo> manuscripts = new ArrayList<ManuscriptInfo>();
    ImageInfo currentImage;
    private Hashtable<String, ImageInfo> hDuplicate = new Hashtable<String, ImageInfo>();
    
    public FixupOCRXML(String inputXMLFile, String outputXMLFile) throws Exception
    {
        this.inputXMLFile = inputXMLFile;
        this.outputXMLFile = outputXMLFile;
    }
    
    
    public void parse () throws Exception {
        // Use the default (non-validating) parser
        SAXParserFactory factory = SAXParserFactory.newInstance();

        // Parse the input
        SAXParser saxParser = factory.newSAXParser();
        // See http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description
        // Uncomment the following to turn on comment processing.
        // http://docs.oracle.com/javaee/1.4/tutorial/doc/JAXPSAX11.html
        // saxParser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
        try {
            File file = new File(inputXMLFile);
            InputStream inputStream= new FileInputStream(file);
            int b;
            int count = 0;
            while (true) {
                b = inputStream.read();
                if (b == '<') {
                    break;
                }
                count++;
            }
            inputStream.close();
            inputStream= new FileInputStream(file);
            for (int i=0; i<count; i++) {
                inputStream.read();
            }
 
            saxParser.parse(inputStream, this);            
            //saxParser.parse( new File(xmlFile), this);
        } catch (Exception ex) {
            throw ex;
        }
    }    

	// Maintains a stack of element names (qName)
    private void push (String qName) {
        pathStack.add(qName);
    }

	// Pop a qName from the stack
    private String pop () {
        int last = pathStack.size()-1;
        String top = null;
        if (last >= 0) {
            top = pathStack.get(last);
            pathStack.remove(last);
        }

        return top;
    }

	// Peek at a qName without popping stack
    private String peek () {
        int last = pathStack.size()-1;
        String top = null;
        if (last >= 0) {
            top = pathStack.get(last);
        }

        return top;

    }

	// Serialize the qName stack. The string that is returned represent the path to the current qName
    private String buildPath () {
        StringBuilder sb = new StringBuilder();
        String delim = "";
        for (String name : pathStack) {
            sb.append(String.format("%s%s", delim, name));
            delim = ".";
        }
        return sb.toString();
    }

    private void displayError(String msg) {
        System.out.format("Line %d: %s\n", locator.getLineNumber(), msg);
    }

    @Override
    public void warning(SAXParseException exception) throws SAXException {
        displayError(exception.getMessage());
    }

    @Override
    public void error(SAXParseException exception) throws SAXException {
        displayError(exception.getMessage());
    }

    @Override
    public void fatalError(SAXParseException exception) throws SAXException {
        displayError(exception.getMessage());
    }

    @Override
    public void setDocumentLocator(Locator locator) {
        this.locator = locator;
    }

    @Override
    public void startDocument() throws SAXException {
    }

    @Override
    public void endDocument() throws SAXException {
        try {
            BufferedWriter bw = IOAssistance.openWriter(outputXMLFile);
            bw.write("<manuscripts>\n");
            for (ManuscriptInfo mi : manuscripts) {
                mi.format(bw);
            }
            bw.write("</manuscripts>\n");
            bw.close();
        } catch (Exception ex) {
            throw new SAXException(ex);            
        }
    }

    @Override
    public void startPrefixMapping(String prefix, String uri) throws SAXException {
    }

    @Override
    public void endPrefixMapping(String prefix) throws SAXException {
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        push(qName);
        if (qName.equals("manuscript")) {
            currentManuscript = new ManuscriptInfo();
            currentManuscript.catalog = atts.getValue("catalog");
            manuscripts.add(currentManuscript);
        } else if (qName.equals("image")) {
            currentJPG = atts.getValue("name");
            ImageInfo ii = hDuplicate.get(currentJPG);
            if (ii != null) {
                currentImage = ii;
            } else {
                currentImage = new ImageInfo(currentJPG);
                hDuplicate.put(currentJPG, currentImage);
                currentManuscript.images.add(currentImage);
            }
        } else if (qName.equals("line")) {
            currentLineInfo = new LineInfo(atts.getValue("n"));
            currentImage.lines.add(currentLineInfo);
            currentLineInfo.corrected = atts.getValue("corrected") != null;
            currentLineInfo.incomplete = atts.getValue("incomplete") != null;
        }
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
        try {
            String path = buildPath();
            StringBuilder buffer = buffers.get(path);
            String value = "";
            if (buffer != null) {
                value = buffer.toString().trim();
                buffer.setLength(0);
            }
            
            if (qName.equals("line")) {
                currentLineInfo.transcription = value;
            } else if (qName.equals("image")) {
            } else if (qName.equals("condition")) {
                currentManuscript.condition = value;
            } else if (qName.equals("manuscript")) {
            }
        } catch (Exception ex) {
            throw new SAXException(ex.getMessage());
        } finally {
            pop();
        }
    }

    private String buildID(String type, int i) {
        return String.format("image.%s.%s%s", currentCropNumber, type, i+1);
    }

    @Override
    // Collects the innerText in a StringBuilder keyed by the current path. It is endElement's responsibility
    // to retrieve the StringBuilder and set its length to 0.
    // This is optional. It is not specified in SAXParserHandler
    public void characters(char[] ch, int start, int length) throws SAXException {
        String path = buildPath();
        StringBuilder sb = buffers.get(path);
        if (sb == null) {
            sb = new StringBuilder();
            buffers.put(path, sb);
        }
        sb.append(new String(ch, start, length));
    }

    @Override
    // Optional if you want to be notified of comments
    public void comment(char[] ch, int start, int length) throws SAXException {

    }
    
    class ManuscriptInfo implements Comparator<LineInfo> {
        String condition;
        String catalog;
        ArrayList<ImageInfo> images = new ArrayList<ImageInfo>();
        
        void format(BufferedWriter bw) throws Exception {
            boolean hasImages = false;
            for (ImageInfo ii : images) {
                if (!ii.jpgName.toLowerCase().contains("unknown")) {
                    hasImages = true;
                    break;
                }
            }
            
            if (hasImages) {
                if (condition == null) {
                    condition = "Condition not written down in TEI file";
                }
                bw.write(String.format(" <manuscript catalog=\"%s\">\n", catalog));
                bw.write(String.format(" <condition>%s</condition>\n", condition));
                
                bw.write(" <images>\n");
                for (ImageInfo ii: images) {
                    if (!ii.jpgName.toLowerCase().contains("unknown")) {
                        Collections.sort(ii.lines, this);
                        LineInfo prev = null;
                        for (LineInfo li : ii.lines) {
                            if (prev != null && prev.n == li.n) {
                                prev.duplicate = true;
                                li.duplicate = true;
                                System.out.format("%s has duplicate line %d\n", ii.jpgName, li.n);
                            }
                            prev = li;
                        }
                        ii.format(bw);
                    }
                }
                bw.write(" </images>\n");
                bw.write(" </manuscript>\n");
            }
        }

        @Override
        public int compare(LineInfo li1, LineInfo li2) {
            return li1.n - li2.n;
        }
    }
    
    
    class ImageInfo {
        String jpgName;
        ArrayList<LineInfo> lines = new ArrayList<LineInfo>(); 
        
        ImageInfo (String jpgName) {
            this.jpgName = jpgName;
        }
        
        void format(BufferedWriter bw) throws Exception {
            bw.write(String.format("  <image name=\"%s\">\n", jpgName));
            for (LineInfo line : lines) {
                line.format(bw);
            }
            bw.write("  </image>\n");
        }
     }
    
    class LineInfo {
        int n;
        boolean corrected = false;
        boolean incomplete = false;
        boolean duplicate = false;
        String transcription;
        LineInfo(String n) {
            this.n = Integer.parseInt(n);
        }
        
        void format(BufferedWriter bw) throws Exception {
            if (transcription != null) {
                bw.write(String.format("   <line n=\"%d\"", n));
                if (incomplete) {
                    bw.write(" incomplete=\"yes\"");
                }
                if (corrected) {
                    bw.write(" corrected=\"yes\"");                
                }
                if (duplicate) {
                    bw.write(" duplicate=\"yes\"");
                }
                bw.write(String.format(">%s</line>\n", transcription));
            }
        }
    }

}
