package qnd;

import java.io.BufferedWriter;
import java.io.File;
import java.util.ArrayList;
import java.util.Hashtable;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.SAXParserAdapter;
import org.sanskritlibrary.XMLUtils;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

public class MDhPFixup extends SAXParserAdapter {
    private String filename;
    private Hashtable<String, ArrayList<SLPInfo>> h  = new Hashtable<String, ArrayList<SLPInfo>>();
    private Hashtable<String, ArrayList<SLPInfo>> slpInfo = null;
    private SLPInfo currentSLP;
    private BufferedWriter out;
    private String currentSID;
    private int currentFullDAtu = 0;
    private String sutraPreds;
    
    MDhPFixup(String filename) {
        this.filename = filename;
    }

    public void parse(Hashtable<String, ArrayList<SLPInfo>> slpInfo) throws Exception {
        SAXParserFactory factory = SAXParserFactory.newInstance();
        this.slpInfo = slpInfo;
        if (slpInfo != null) {
            out = IOAssistance.openWriter("temp.xml");
        }
        //makeGroupComplete("^11(ab)\\(cd\\)(?ef)gh(yz)22$", new Leaf("", ""));
        /*
         SchemaFactory schemaFactory = SchemaFactory.newInstance(XMLConstants.XML_DTD_NS_URI);
         Schema schema = schemaFactory.newSchema(new File("files/rules.dtd"));
         Validator validator = schema.newValidator();
         validator.validate(new StreamSource("files/rules.xml"));
         */

        SAXParser saxParser = factory.newSAXParser();

        // See http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description
        // Uncomment the following to turn on comment processing.
        // http://docs.oracle.com/javaee/1.4/tutorial/doc/JAXPSAX11.html
        // saxParser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
        try {
            saxParser.parse(new File(filename), this);
        } catch (Exception ex) {
            System.out.format("%d. %s: %s\n", locator.getLineNumber(), filename, ex.toString());
        }
    }
    
    
    public Hashtable<String, ArrayList<SLPInfo>> getSLPInfo () {
        return h;
    }
    
    @Override
    public void startDocument () throws SAXException {
        try {
        } catch (Exception ex) {
            throw new SAXException(ex);
        }
    }
    
    @Override
    public void endDocument() throws SAXException {
        try {
            if (slpInfo != null) {
                out.close();
                XMLUtils xu = new XMLUtils(XMLUtils.Types.FORMAT, "temp.xml", "out.xml");
                xu.parse();
            }
            
        } catch (Throwable ex) {
            throw new SAXException(ex.getMessage());
        }
    }
    
    
    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        push(qName, atts);
        if (locator.getLineNumber() == 438) {
            localName = null;
        }
        
        if (slpInfo == null) {
            if (qName.equals("sUtra")) {
                currentSID = atts.getValue("sid");
                sutraPreds = atts.getValue("preds");
                if (h.get(currentSID) != null) {
                    System.out.format("Duplicate sid: %s\n", currentSID);
                }
                h.put(currentSID, new ArrayList<SLPInfo>());
            } else if (qName.equals("entry")) {
                String preds = atts.getValue("preds");
                if (preds == null) {
                    if (sutraPreds != null && !sutraPreds.equals("none")) {
                        String path = buildPath();
                        if (!path.contains("pAWAntara")) {
                            System.out.format("sutraPreds=%s\n", sutraPreds);
                        }
                    }
                    preds = "none";
                }
                currentSLP = new SLPInfo();
                String entryID = atts.getValue("id");
                currentSLP.sid = String.format("%s[%s]", currentSID, entryID);
                currentSLP.preds = preds;
                ArrayList<SLPInfo> a = h.get(currentSID);
                a.add(currentSLP);
            }
        } else {
            try {
                out.write(String.format("<%s ", qName));
                int numatts = atts.getLength();
                for (int i=0; i<numatts; i++) {
                    String name = atts.getQName(i);
                    String value = atts.getValue(i);
                    if (qName.equals("sUtra") && name.equals("sid")) {
                        currentSID = value;
                        currentFullDAtu = 0;
                    }
                    out.write(String.format("%s=\"%s\" ", name, value));
                }
                
                if (qName.equals("fullDAtu")) {
                    ArrayList<SLPInfo> a = slpInfo.get(currentSID);
                    if (a.size()>0) {
                        SLPInfo si = a.get(currentFullDAtu++);
                        if (si != null) {
                            out.write(String.format("slp=\"%s\"", si.slp));
                        }
                    } else {
                        out.write("slp=\"*\"");
                    }
                }
                
                out.write(">");
            } catch (Exception ex) {
                throw new SAXException(ex);
            }
        }
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
        String path = buildPath();
        String value = "";
        StringBuilder sb = buffers.get(path);
        if (sb != null) {
            value = sb.toString().trim();
            sb.setLength(0);
        }
        
        if (slpInfo == null) {
            if (currentSLP != null) {
                if (qName.equals("root")) {
                    currentSLP.root = value;
                } else if (qName.equals("marker")) {
                    currentSLP.marker = value;
                } else if (qName.equals("premarker")) {
                    currentSLP.premarker = value;
                } else if (qName.equals("entry")) {
                    currentSLP.computeSLP();
                    currentSLP = null;
                }
            }
        } else {
            try {
                out.write(value);
                out.write(String.format("</%s>", qName));
            } catch (Exception ex) {
                throw new SAXException(ex);
            }
        }
                
        pop();
    }
    
    @Override
    public void characters(char[] ch, int start, int length) throws SAXException {
        String path = buildPath();
        StringBuilder buffer = buffers.get(path);
        if (buffer == null) {
            buffer = new StringBuilder();
            buffers.put(path, buffer);
        }
        buffer.append(new String(ch, start, length));
    }    
}
