package qnd;

import java.io.BufferedReader;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import org.sanskritlibrary.IOAssistance;

public class OCRPageSelection {

    String manuscript;
    int numIterations;
    ArrayList<LineInfo> lines = new ArrayList<LineInfo>();
    Hashtable<String, SyllableInfo> glyphsHash = new Hashtable<String, SyllableInfo>();
    StringBuilder sbWorkingSyllable = new StringBuilder();
    private final static int SLP_vowel = 1;
    private final static int SLP_consonent = 2;
    private final static int SLP_atom = 4;
    private final static int SLP_accent = 8;
    private final static int SLP_whitespace = 16;
    private final static int SLP_conjunct = 32;
    SyllableInfo[] baseSLP = new SyllableInfo[]{
        new SyllableInfo("a", SLP_vowel),
        new SyllableInfo("A", SLP_vowel),
        new SyllableInfo("i", SLP_vowel),
        new SyllableInfo("I", SLP_vowel),
        new SyllableInfo("u", SLP_vowel),
        new SyllableInfo("U", SLP_vowel),
        new SyllableInfo("f", SLP_vowel),
        new SyllableInfo("F", SLP_vowel),
        new SyllableInfo("x", SLP_vowel),
        new SyllableInfo("X", SLP_vowel),
        new SyllableInfo("e", SLP_vowel),
        new SyllableInfo("E", SLP_vowel),
        new SyllableInfo("o", SLP_vowel),
        new SyllableInfo("O", SLP_vowel),
        new SyllableInfo("h", SLP_consonent),
        new SyllableInfo("H", SLP_vowel),
        new SyllableInfo("m", SLP_consonent),
        new SyllableInfo("M", SLP_vowel),
        new SyllableInfo("k", SLP_consonent),
        new SyllableInfo("K", SLP_consonent),
        new SyllableInfo("g", SLP_consonent),
        new SyllableInfo("G", SLP_consonent),
        new SyllableInfo("n", SLP_consonent),
        new SyllableInfo("N", SLP_consonent),
        new SyllableInfo("c", SLP_consonent),
        new SyllableInfo("C", SLP_consonent),
        new SyllableInfo("j", SLP_consonent),
        new SyllableInfo("J", SLP_consonent),
        new SyllableInfo("Y", SLP_consonent),
        new SyllableInfo("w", SLP_consonent),
        new SyllableInfo("W", SLP_consonent),
        new SyllableInfo("q", SLP_consonent),
        new SyllableInfo("Q", SLP_consonent),
        new SyllableInfo("R", SLP_consonent),
        new SyllableInfo("t", SLP_consonent),
        new SyllableInfo("T", SLP_consonent),
        new SyllableInfo("d", SLP_consonent),
        new SyllableInfo("D", SLP_consonent),
        new SyllableInfo("p", SLP_consonent),
        new SyllableInfo("P", SLP_consonent),
        new SyllableInfo("b", SLP_consonent),
        new SyllableInfo("B", SLP_consonent),
        new SyllableInfo("y", SLP_consonent),
        new SyllableInfo("r", SLP_consonent),
        new SyllableInfo("l", SLP_consonent),
        new SyllableInfo("v", SLP_consonent),
        new SyllableInfo("s", SLP_consonent),
        new SyllableInfo("S", SLP_consonent),
        new SyllableInfo("z", SLP_consonent),
        new SyllableInfo("L", SLP_consonent),
        new SyllableInfo("|", SLP_consonent),
        new SyllableInfo("'", SLP_atom),
        new SyllableInfo(";", SLP_atom),
        new SyllableInfo("-", SLP_atom),
        new SyllableInfo("~", SLP_accent),
        new SyllableInfo("\\", SLP_accent),
        new SyllableInfo(" ", SLP_whitespace),
        new SyllableInfo("\t", SLP_whitespace),
        new SyllableInfo("\r", SLP_whitespace),
        new SyllableInfo("\n", SLP_whitespace),
        new SyllableInfo(".", SLP_atom),
        new SyllableInfo("0", SLP_atom),
        new SyllableInfo("1", SLP_atom),
        new SyllableInfo("2", SLP_atom),
        new SyllableInfo("3", SLP_atom),
        new SyllableInfo("4", SLP_atom),
        new SyllableInfo("5", SLP_atom),
        new SyllableInfo("6", SLP_atom),
        new SyllableInfo("7", SLP_atom),
        new SyllableInfo("8", SLP_atom),
        new SyllableInfo("9", SLP_atom)
    };

    OCRPageSelection(String manuscript, int numIterations) {
        this.manuscript = manuscript;
        this.numIterations = numIterations;
        initializeGlyphHash();
    }

    private void dumpSLPInfo() {
        SyllableInfo[] a = new SyllableInfo[glyphsHash.size()];
        Enumeration<String> en = glyphsHash.keys();
        int i = 0;
        while (en.hasMoreElements()) {
            String key = en.nextElement();
            a[i++] = glyphsHash.get(key);
        }

        boolean sorted = false;
        int j = baseSLP.length - 1;

        while (!sorted) {
            sorted = true;
            for (i = 0; i < j; i++) {
                //if (a[i].count < a[i+1].count) {
                if (a[i].glyph.compareTo(a[i + 1].glyph) > 0) {
                    SyllableInfo swap = a[i];
                    a[i] = a[i + 1];
                    a[i + 1] = swap;
                    sorted = false;
                }
            }
            j = j - 1;
        }

        for (i = 0; i < a.length; i++) {
            if (a[i].glyph.length() > 1) {
                System.out.format("%s %d (%s)\n", a[i].glyph, a[i].count, a[i].markConjunct());
            }
        }
    }

    private void initializeGlyphHash() {
        for (SyllableInfo si : baseSLP) {
            glyphsHash.put(si.glyph, si);
        }
    }

    public int execute() throws Exception {

        BufferedReader br = IOAssistance.openReader(manuscript);
        String line;
        while ((line = br.readLine()) != null) {
            lines.add(new LineInfo(line));
        }
        br.close();

        findAllGlyphs();
        dumpSLPInfo();
        return doIterations();
    }

    private SegmentInfo[] segments = new SegmentInfo[] {
new SegmentInfo("0002", 1, 63),
new SegmentInfo("0003", 64, 135),
new SegmentInfo("0004", 136, 190),
new SegmentInfo("0005", 191, 251),
new SegmentInfo("0006", 252, 314),
new SegmentInfo("0007", 315, 382),
new SegmentInfo("0008", 383, 443),
new SegmentInfo("0009", 444, 502),
new SegmentInfo("0010", 503, 567),
new SegmentInfo("0011", 568, 628),
new SegmentInfo("0012", 629, 683),
new SegmentInfo("0013", 684, 735),
new SegmentInfo("0014", 736, 788),
new SegmentInfo("0015", 789, 841),
new SegmentInfo("0016", 842, 899),
new SegmentInfo("0017", 900, 955),
new SegmentInfo("0018", 956, 1012),
new SegmentInfo("0019", 1013, 1068),
new SegmentInfo("0020", 1069, 1126),
new SegmentInfo("0021", 1127, 1184),
new SegmentInfo("0022", 1185, 1235),
new SegmentInfo("0023", 1236, 1294),
new SegmentInfo("0024", 1295, 1348),
new SegmentInfo("0025", 1349, 1395),
new SegmentInfo("0026", 1396, 1442),
new SegmentInfo("0027", 1443, 1494),
new SegmentInfo("0028", 1495, 1547),
new SegmentInfo("0029", 1548, 1604),
new SegmentInfo("0030", 1605, 1657),
new SegmentInfo("0031", 1658, 1712),
new SegmentInfo("0032", 1713, 1763),
new SegmentInfo("0033", 1764, 1820),
new SegmentInfo("0034", 1821, 1876),
new SegmentInfo("0035", 1877, 1930),
new SegmentInfo("0036", 1931, 1980),
new SegmentInfo("0037", 1981, 2033),
new SegmentInfo("0038", 2034, 2083),
new SegmentInfo("0039", 2084, 2135),
new SegmentInfo("0040", 2136, 2188),
new SegmentInfo("0041", 2189, 2239),
new SegmentInfo("0042", 2240, 2289),
new SegmentInfo("0043", 2290, 2332),
new SegmentInfo("0044", 2333, 2380),
new SegmentInfo("0045", 2381, 2424),
new SegmentInfo("0046", 2425, 2466),
new SegmentInfo("0047", 2467, 2511),
new SegmentInfo("0048", 2512, 2555),
new SegmentInfo("0049", 2556, 2596),
new SegmentInfo("0050", 2597, 2635),
new SegmentInfo("0051", 2636, 2679),
new SegmentInfo("0052", 2680, 2720),
new SegmentInfo("0053", 2721, 2762),
new SegmentInfo("0054", 2763, 2809),
new SegmentInfo("0055", 2810, 2851),
new SegmentInfo("0056", 2852, 2896),
new SegmentInfo("0057", 2897, 2947),
new SegmentInfo("0058", 2948, 3001)        
    };
    private int doIterations() {
        ArrayList<SyllableInfo> contenders = new ArrayList<SyllableInfo>();
        int numLines = lines.size();
        int bestSegment = 0;
        int numSegments = segments.length;

        boolean moreSyllables = true;
        for (int i = 0; i < numIterations && moreSyllables; i++) {
            int maxGlyphs = 0;          // each iteration computes the segment with the most glyphs
            for (int k = 0; k < numSegments; k++) {

                if (!segments[k].active) {          // this segment was the max-glyph segment in a previous iteration.
                    continue;
                }

                int first = segments[k].start;
                int last = segments[k].end;
                int glyphCount = 0;
                for (int n = first; n <= last; n++) {
                    LineInfo li = lines.get(n);
                    ArrayList<String> syllables = getSyllables(li.text);
                    for (String glyph : syllables) {
                        SyllableInfo si = glyphsHash.get(glyph);
                        if (si.active) {
                            if (!si.current) {
                                glyphCount++;
                                si.current = true;
                            }
                        } else {
                            si.active = false;
                        }
                    }
                }
                // reach here after counting glyphs in current block of iterationSize lines
                if (glyphCount > maxGlyphs) {
                    maxGlyphs = glyphCount;
                    bestSegment = k;
                    contenders.clear();
                }
                
                // reset all current fields to false for next segment
                Enumeration<String> en = glyphsHash.keys();
                while (en.hasMoreElements()) {
                    String key = en.nextElement();
                    SyllableInfo si = glyphsHash.get(key);
                    if (si.current) {
                        if (bestSegment == k) {
                            contenders.add(si);         // remember glyphs collected for this segment if it is the best so far
                        }
                        si.current = false;
                    }
                }
            }   // end of active segments loop

            segments[bestSegment].active = false;
            for (SyllableInfo si : contenders) {
                si.active = false;          // won't be available for the next iteration
            }
            
            // see if there are any more glyphs to find
            moreSyllables = false;
            Enumeration<String> en = glyphsHash.keys();
            while (!moreSyllables && en.hasMoreElements()) {
                String key = en.nextElement();
                SyllableInfo si = glyphsHash.get(key);
                if (si.active) {
                    moreSyllables = true;
                }
            }
            
        } // end of iterations loop
        // Reach here when iterations complete
        Enumeration<String> en = glyphsHash.keys();
        int cnt = 0;
        double total = 0;
        int max = 0;
        while (en.hasMoreElements()) {
            String key = en.nextElement();
            SyllableInfo si = glyphsHash.get(key);
            if (si.active) {
                total += si.count;
                cnt++;
                if (si.count > max) {
                    max = si.count;
                }
            }
        }
        int totalLines = 0;
        for (SegmentInfo si : segments) {
            if (!si.active) {
                totalLines += si.end - si.start + 1;
            }
        }
        
        System.out.format("#iteration:%d, #numPending:%d of %d, max:%d, avg:%s #lines:%d\n",
                numIterations, cnt, glyphsHash.size(), max, Double.toString(total/cnt), totalLines);
        for (int i=0; i<segments.length; i++) {
            if (!segments[i].active) {
                System.out.format("\t%s: %d-%d\n", segments[i].image, segments[i].start, segments[i].end);
            }
        }
        return totalLines;
    }

    private boolean saveVowelWithConjunct = false;
    private ArrayList<String> getSyllables(String line) {
        int lastType = 0;
        ArrayList<String> syllables = new ArrayList<String>();
        sbWorkingSyllable.setLength(0);
        int len = line.length();
        for (int i = 0; i < len; i++) {
            String ch = "" + line.charAt(i);
            SyllableInfo currentChar = glyphsHash.get(ch);
            if (currentChar == null) {
                System.out.format("%s not found\n", ch);
                continue;
            } else {
                int type = currentChar.type;
                boolean appendCurrentCharacter = true;
                if (currentChar.type == SLP_consonent) {
                    if (lastType != SLP_consonent) {
                        saveSyllable(syllables, sbWorkingSyllable.toString());
                    }
                } else if (sbWorkingSyllable.length() > 0) {                    // current char is not a consonent
                    if (currentChar.type == SLP_vowel) {
                        if (saveVowelWithConjunct) {
                            if ((lastType & (SLP_vowel | SLP_accent)) != 0) {
                                saveSyllable(syllables, sbWorkingSyllable.toString());   // have consecutive vowels
                            }
                        } else {
                            saveSyllable(syllables, sbWorkingSyllable.toString());
                        }
                    } else if (currentChar.type == SLP_atom) {
                        saveSyllable(syllables, sbWorkingSyllable.toString());
                        lastType = SLP_whitespace;          // treat atoms like whitespace
                    } else if (currentChar.type == SLP_whitespace) {
                        if (lastType != SLP_consonent) {    // if consonent append what follows white space.
                            saveSyllable(syllables, sbWorkingSyllable.toString());
                        } else {
                            type = SLP_consonent;               // make it appear that blank did not exist                                
                        }
                        appendCurrentCharacter = false;         // don't save white space
                    }
                }
                lastType = type;
                if (appendCurrentCharacter) {
                    sbWorkingSyllable.append(currentChar.glyph);
                }
            }
        }
        return syllables;
    }
    
    

    private void findAllGlyphs() {
        for (LineInfo line : lines) {
            ArrayList<String> glyphs = getSyllables(line.text);
            for (String glyph : glyphs) {
                SyllableInfo slp = glyphsHash.get(glyph);
                if (slp == null) {
                    slp = new SyllableInfo(glyph, SLP_conjunct);
                    slp.line = line.text;
                    glyphsHash.put(glyph, slp);
                }
                slp.count++;
            }
        }  // end iteration through lines        
    }

    private void saveSyllable(ArrayList<String> glyphs, String glyph) {
        glyphs.add(glyph);
        sbWorkingSyllable.setLength(0);
    }

    class LineInfo {

        String text;
        boolean active;

        LineInfo(String text) {
            this.text = text;
            active = true;
        }
    }

    class SyllableInfo {

        String glyph;
        String line;
        int type;
        int count;
        boolean active;                     // if glyph should be used in current iteration
        boolean current;                    // if glyph found in current iteration
        int iterationNumber = -1;           // if other than -1 then this is the iteration that added this glyph.

        SyllableInfo(String glyph, int type) {
            this.glyph = glyph;
            this.type = type;
            this.active = true;
        }

        String markConjunct() {
            StringBuilder sb = new StringBuilder();
            if (line != null) {
                if (line.contains(glyph)) {
                    sb.append(line.replace(glyph, String.format("[%s]", glyph)));
                } else {
                    ArrayList<String> permutations = getPermutations();
                    for (String permutation : permutations) {
                        if (line.contains(permutation)) {
                            sb.append(line.replace(permutation, String.format("[%s]", permutation)));
                            break;
                        }
                    }
                }
            }

            return sb.toString();
        }

        ArrayList<String> getPermutations() {
            ArrayList<String> a = new ArrayList<String>();
            StringBuilder sb = new StringBuilder();
            char ch[] = new char[glyph.length()];
            for (int i = 0; i < ch.length; i++) {
                ch[i] = glyph.charAt(i);
            }
            int blankLocation = 1;
            for (int i = 0; i < ch.length - 1; i++) {
                for (int j = 0; j < ch.length; j++) {
                    if (j == blankLocation) {
                        sb.append(" ");
                    }
                    sb.append(ch[j]);
                }
                a.add(sb.toString());
                sb.setLength(0);
                blankLocation++;
            }

            return a;
        }
    }

    class SegmentInfo {

        int start;
        int end;
        String image;
        boolean active;

        SegmentInfo(String image, int start, int end) {
            this.image = image;
            this.start = start;
            this.end = end;
            active = true;
        }
    }
}
