package qnd;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Hashtable;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.Regex;
import org.sanskritlibrary.SAXParserAdapter;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;


// "/Users/ralph/SanskritLibrary/data/mss/UPennCropped/mscoll390_item349_wk1/mscoll390_item349_wk1_body0002_crop1.jpg"
public class OCRPrep extends SAXParserAdapter {
    StringBuilder sb = new StringBuilder();
    static Regex reLine = new Regex("", "^([*x]{0,2})Line\\s+(\\d+):(.*)");
    static StringBuilder copyScript = new StringBuilder();
    static StringBuilder htmlScript = new StringBuilder("<html><head><title>Images</title></head><body>\n");
    static StringBuilder errors = new StringBuilder();
    static Hashtable<String, String> textElements = new Hashtable<String, String>();
    Hashtable<String, String> folios = new Hashtable<String, String>();
    static final Hashtable<String, String> perm = new Hashtable<String, String>();
    String filename;
    String shortName;
    String manuscriptDirectory = "/Users/ralph/SanskritLibrary/data/mss/";
    ArrayList<Boolean> permStack = new ArrayList<Boolean>();
    
    static final Hashtable<String, String> dirmap = new Hashtable<String, String>();
    static {
        perm.put("ab", "allow"); 		// so child elements can contribute (probably gap)
//add: penn0349:986        ignore
//bibl: penn0349:996       ignore
//biblScope: penn0349:996  ignore
        perm.put("body", "allow");      // allow so child elements can contribute
        perm.put("choice", "??");     	// insert "??" and ignore
//corr: penn0349:997       ignore
        perm.put("del", "allow");		// allow (assuming it is visible and tell OCR it is the stuff that is deleted)
//desc: penn0349:1023      ignore
        perm.put("div", "allow"); 		// allow so child elements can contribute
//foreign: penn0375:4105   ignore
        perm.put("gap", "..");			// insert ..
        perm.put("head", "allow"); 		// allow so child elements can contribute
//item: penn2376:2003      ignore (assume it is part of list)
        perm.put("l", "allow");			// allow so child elements can contribute
        perm.put("lb", "break");		// insert cr followed by Line n: 
        perm.put("lg", "allow");        // allow so child elements can contribute
//list: penn2376:2001      ignore
//locus: penn2247:5252     ignore
//note: penn0351:1443      ignore
        perm.put("p", "allow");			// allow so child elements can contribute
//persName: penn2249:1516  ignore
        perm.put("quote", "allow"); 	// allow so child elements can contribute
// ref: penn0375:4190       ignore
// reg: penn2173:1106       ignore (contains a reading that has been regularized in some sense)
        perm.put("s", "allow");     	// allow so child elements can contribute
        perm.put("seg", "allow");       // allow so child elements can contribute
//sic: penn0349:997        ignore
        perm.put("sp", "allow");        // allow so child elements can contribute
        perm.put("speaker", "allow");	// allow so child elements can contribute
//supplied: penn0349:1020  ignore
        perm.put("text", "allow"); 		// allow so child elements can contribute
        perm.put("title", "allow");		// allow (but seems to be child of ignored bibl element)
        perm.put("trailer", "allow");	// allow so child elements can contribute
        perm.put("unclear", "??");		// allow but is probably child of choice element


        dirmap.put("br32", "BrownCropped/32");
        dirmap.put("br33", "BrownCropped/09.34_item_33_TIFF");
        dirmap.put("br34", "BrownCropped/09.34_item_34_TIFF");
        dirmap.put("br35", "BrownCropped/09.34_item_35_TIFF");
		dirmap.put("br36", "BrownCropped/36");
		dirmap.put("br37", "BrownCropped/37");
		dirmap.put("br38", "BrownCropped/38");
		dirmap.put("br44", "BrownCropped/44");
		dirmap.put("br45", "BrownCropped/45");
		dirmap.put("br46", "BrownCropped/46");
		dirmap.put("br47", "BrownCropped/47");
		dirmap.put("br49", "BrownCropped/49");
		dirmap.put("br50", "BrownCropped/50");
		dirmap.put("br52", "BrownCropped/52");
		dirmap.put("penn0349", "UPennCropped/mscoll390_item349_wk1");
		dirmap.put("penn0351", "UPennCropped/mscoll390_item351");
		dirmap.put("penn0375", "UPennCropped/mscoll390_item375");
		dirmap.put("penn0388", "UPennCropped/mscoll390_item388");
		dirmap.put("penn0390", "UPennCropped/mscoll390_item390");
		dirmap.put("penn0391", "UPennCropped/mscoll390_item391_wk1");
		dirmap.put("penn0398", "UPennCropped/mscoll390_item398");
		dirmap.put("penn0401", "UPennCropped/mscoll390_item401");
		dirmap.put("penn0426", "UPennCropped/mscoll390_item426");
		dirmap.put("penn0448", "UPennCropped/mscoll390_item448");
		dirmap.put("penn0452", "UPennCropped/mscoll390_item452_wk1");
		dirmap.put("penn0488", "UPennCropped/mscoll390_item488");
		dirmap.put("penn0489", "UPennCropped/mscoll390_item489");
		dirmap.put("penn0490", "UPennCropped/mscoll390_item490");
		dirmap.put("penn0491", "UPennCropped/mscoll390_item491");
		dirmap.put("penn0492", "UPennCropped/mscoll390_item492_wk1");
		dirmap.put("penn0515", "UPennCropped/mscoll390_item515_wk1");
		dirmap.put("penn0555", "UPennCropped/mscoll390_item555_wk1");
		dirmap.put("penn0559", "UPennCropped/mscoll390_item559");
		dirmap.put("penn0569", "UPennCropped/mscoll390_item569");
		dirmap.put("penn0749", "UPennCropped/mscoll390_item749_wk1");
		dirmap.put("penn0773", "UPennCropped/mscoll390_item773");
		dirmap.put("penn0906", "UPennCropped/mscoll390_item906");
		dirmap.put("penn1975", "UPennCropped/mscoll390_item1975_wk1");
		dirmap.put("penn2173", "UPennCropped/mscoll390_item2173");
		dirmap.put("penn2174", "UPennCropped/mscoll390_item2174_wk1");
		dirmap.put("penn2175", "UPennCropped/mscoll390_item2175");
		dirmap.put("penn2180", "UPennCropped/mscoll390_item2180_wk1");
		dirmap.put("penn2182", "UPennCropped/mscoll390_item2182");
		dirmap.put("penn2184", "UPennCropped/mscoll390_item2184_wk1");
		dirmap.put("penn2186", "UPennCropped/mscoll390_item2186");
		dirmap.put("penn2197", "UPennCropped/mscoll390_item2197_wk1");
		dirmap.put("penn2198", "UPennCropped/mscoll390_item2198_wk1");
		dirmap.put("penn2199", "UPennCropped/mscoll390_item2199");
		dirmap.put("penn2202", "UPennCropped/mscoll390_item2202_wk1");
		dirmap.put("penn2222", "UPennCropped/mscoll390_item2222");
		dirmap.put("penn2228", "UPennCropped/mscoll390_item2228_wk1");
		dirmap.put("penn2231", "UPennCropped/mscoll390_item2231");
		dirmap.put("penn2233", "UPennCropped/mscoll390_item2233");
		dirmap.put("penn2239", "UPennCropped/mscoll390_item2239_wk1");
		dirmap.put("penn2241", "UPennCropped/mscoll390_item2241_wk1");
		dirmap.put("penn2242", "UPennCropped/mscoll390_item2242_wk1");
		dirmap.put("penn2247", "UPennCropped/mscoll390_item2247");
		dirmap.put("penn2248", "UPennCropped/mscoll390_item2248_wk1");
		dirmap.put("penn2249", "UPennCropped/mscoll390_item2249");
		dirmap.put("penn2250", "UPennCropped/mscoll390_item2250_wk1");
		dirmap.put("penn2251", "UPennCropped/mscoll390_item2251");
		dirmap.put("penn2252", "UPennCropped/mscoll390_item2252_wk1");
		dirmap.put("penn2253", "UPennCropped/mscoll390_item2253_wk1");
		dirmap.put("penn2254", "UPennCropped/mscoll390_item2254_wk1");
		dirmap.put("penn2259", "UPennCropped/mscoll390_item2259_wk1");
		dirmap.put("penn2260", "UPennCropped/mscoll390_item2260_wk1");
		dirmap.put("penn2279", "UPennCropped/mscoll390_item2279");
		dirmap.put("penn2298", "UPennCropped/mscoll390_item2298_wk1");
		dirmap.put("penn2299", "UPennCropped/mscoll390_item2299_wk1");
		dirmap.put("penn2300", "UPennCropped/mscoll390_item2300_wk1");
		dirmap.put("penn2301", "UPennCropped/mscoll390_item2301_wk1");
		dirmap.put("penn2302", "UPennCropped/mscoll390_item2302_wk1");
		dirmap.put("penn2303", "UPennCropped/mscoll390_item2303");
		dirmap.put("penn2304", "UPennCropped/mscoll390_item2304_wk1");
		dirmap.put("penn2305", "UPennCropped/mscoll390_item2305");
		dirmap.put("penn2310", "UPennCropped/mscoll390_item2310_wk1");
		dirmap.put("penn2311", "UPennCropped/mscoll390_item2311");
		dirmap.put("penn2326", "UPennCropped/mscoll390_item2326");
		dirmap.put("penn2327", "UPennCropped/mscoll390_item2327");
		dirmap.put("penn2330", "UPennCropped/mscoll390_item2330");
		dirmap.put("penn2334", "UPennCropped/mscoll390_item2334");
		dirmap.put("penn2335", "UPennCropped/mscoll390_item2335_wk1");
		dirmap.put("penn2336", "UPennCropped/mscoll390_item2336");
		dirmap.put("penn2339", "UPennCropped/mscoll390_item2339");
		dirmap.put("penn2340", "UPennCropped/mscoll390_item2340_wk1");
		dirmap.put("penn2341", "UPennCropped/mscoll390_item2341_wk1");
		dirmap.put("penn2343", "UPennCropped/mscoll390_item2343");
		dirmap.put("penn2348", "UPennCropped/mscoll390_item2348");
		dirmap.put("penn2352", "UPennCropped/mscoll390_item2352_wk1");
		dirmap.put("penn2363", "UPennCropped/mscoll390_item2363");
		dirmap.put("penn2366", "UPennCropped/mscoll390_item2366_wk1");
		dirmap.put("penn2367", "UPennCropped/mscoll390_item2367_wk1");
		dirmap.put("penn2368", "UPennCropped/mscoll390_item2368_wk1");
		dirmap.put("penn2369", "UPennCropped/mscoll390_item2369_wk1");
		dirmap.put("penn2370", "UPennCropped/mscoll390_item2370");
		dirmap.put("penn2372", "UPennCropped/mscoll390_item2372_wk1");
		dirmap.put("penn2373", "UPennCropped/mscoll390_item2373_wk1");
		dirmap.put("penn2374", "UPennCropped/mscoll390_item2374_wk1");
		dirmap.put("penn2375", "UPennCropped/mscoll390_item2375_wk1");
		dirmap.put("penn2376", "UPennCropped/mscoll390_item2376_wk1");
		dirmap.put("penn2377", "UPennCropped/mscoll390_item2377_wk1");
		dirmap.put("penn2378", "UPennCropped/mscoll390_item2378_wk1");
		dirmap.put("penn2379", "UPennCropped/mscoll390_item2379_wk1");
		dirmap.put("penn2380", "UPennCropped/mscoll390_item2380_wk1");
		dirmap.put("penn2381", "UPennCropped/mscoll390_item2381_wk1");
		dirmap.put("penn2382", "UPennCropped/mscoll390_item2382_wk1");
		dirmap.put("penn2383", "UPennCropped/mscoll390_item2383_wk1");
		dirmap.put("penn2384", "UPennCropped/mscoll390_item2384_wk1");
		dirmap.put("penn2385", "UPennCropped/mscoll390_item2385_wk1");
		dirmap.put("penn2388", "UPennCropped/mscoll390_item2388");
		dirmap.put("penn2390", "UPennCropped/mscoll390_item2390_wk1");
		dirmap.put("penn2396", "UPennCropped/mscoll390_item2396_wk1");
		dirmap.put("penn2397", "UPennCropped/mscoll390_item2397_wk1");
		dirmap.put("penn2398", "UPennCropped/mscoll390_item2398");
		dirmap.put("penn2402", "UPennCropped/mscoll390_item2402_wk1");
		dirmap.put("penn2408", "UPennCropped/mscoll390_item2408");
		dirmap.put("penn2412", "UPennCropped/mscoll390_item2412");
		dirmap.put("penn2437", "UPennCropped/mscoll390_item2437_wk1");
		dirmap.put("penn2445", "UPennCropped/mscoll390_item2445");
		dirmap.put("penn2453", "UPennCropped/mscoll390_item2453");
		dirmap.put("penn2462", "UPennCropped/mscoll390_item2462");
		dirmap.put("penn2463", "UPennCropped/mscoll390_item2463");
		dirmap.put("penn2464", "UPennCropped/mscoll390_item2464_wk1");
		dirmap.put("penn2469", "UPennCropped/mscoll390_item2469_wk1");
		dirmap.put("penn2475", "UPennCropped/mscoll390_item2475");
		dirmap.put("penn2476", "UPennCropped/mscoll390_item2476");
		dirmap.put("penn2486", "UPennCropped/mscoll390_item2486");
		dirmap.put("penn2487", "UPennCropped/mscoll390_item2487");
		dirmap.put("penn2491", "UPennCropped/mscoll390_item2491");
		dirmap.put("penn2496", "UPennCropped/mscoll390_item2496_wk1");
		dirmap.put("penn2514", "UPennCropped/mscoll390_item2514");
		dirmap.put("penn2563", "UPennCropped/mscoll390_item2563_wk1");
		dirmap.put("penn2574", "UPennCropped/mscoll390_item2574");
		dirmap.put("penn2579", "UPennCropped/mscoll390_item2579_wk1");
		dirmap.put("penn2609", "UPennCropped/mscoll390_item2609_wk1");
		dirmap.put("penn2617", "UPennCropped/mscoll390_item2617_wk1");
		dirmap.put("penn2618", "UPennCropped/mscoll390_item2618_wk1");
		dirmap.put("penn2619", "UPennCropped/mscoll390_item2619");
		dirmap.put("penn2620", "UPennCropped/mscoll390_item2620");
		dirmap.put("penn2621", "UPennCropped/mscoll390_item2621");
		dirmap.put("penn2622", "UPennCropped/mscoll390_item2622_wk1");
		dirmap.put("penn2623", "UPennCropped/mscoll390_item2623");
		dirmap.put("penn2624", "UPennCropped/mscoll390_item2624_wk1");
		dirmap.put("penn2625", "UPennCropped/mscoll390_item2625");
		dirmap.put("penn2626", "UPennCropped/mscoll390_item2626_wk1");
		dirmap.put("penn2627", "UPennCropped/mscoll390_item2627");
		dirmap.put("penn2628", "UPennCropped/mscoll390_item2628");
		dirmap.put("penn2629", "UPennCropped/mscoll390_item2629");
		dirmap.put("penn2639", "UPennCropped/mscoll390_item2639_wk1");
		dirmap.put("penn2654", "UPennCropped/mscoll390_item2654_wk1");
		dirmap.put("penn2664", "UPennCropped/mscoll390_item2664_wk1");
		dirmap.put("penn2666", "UPennCropped/mscoll390_item2666_wk1");
		dirmap.put("penn2669", "UPennCropped/mscoll390_item2669");
		dirmap.put("penn2688", "UPennCropped/mscoll390_item2688_wk1");
		dirmap.put("penn2780", "UPennCropped/mscoll390_item2780");
		dirmap.put("penn2788", "UPennCropped/mscoll390_item2788");
		dirmap.put("penn2814", "UPennCropped/mscoll390_item2814_wk1");
		dirmap.put("penn2815", "UPennCropped/mscoll390_item2815_wk1");
		dirmap.put("msindic5", "UPennCropped/msindic5");        
    }
    
    OCRPrep(String manuscriptDirectory, String filename, String shortName) throws Exception {
        this.filename = filename;
        this.shortName = shortName.substring(0, shortName.length() - 4);
        this.manuscriptDirectory = manuscriptDirectory;
        String subdir = dirmap.get(this.shortName);
        if (subdir == null) {
            System.out.format("%s not found.", this.shortName);
        } else {
            this.manuscriptDirectory += String.format("%s", subdir);
            SAXParserFactory factory = SAXParserFactory.newInstance();

            // Parse the input
            SAXParser saxParser = factory.newSAXParser();
            // See http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description
            // Uncomment the following to turn on comment processing.
            // http://docs.oracle.com/javaee/1.4/tutorial/doc/JAXPSAX11.html
            // saxParser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
            try {
                saxParser.parse(new File(filename), this);
            } catch (Exception ex) {
                throw ex;
            }
        }
    }
    
    public String getDescription () {
        return sb.toString();
    }
    
    public static String getCopyScript () {
        return copyScript.toString();
    }
    
    public static String getHTMLScript () {
        return htmlScript.toString();
    }
    
    
    @Override
    public void startDocument () {
        copyScript.append("\n");
        htmlScript.append(String.format("<h3>%s</h3>\n", shortName));
    }
    
    public static void dumpElements () {
        ArrayList<String> a = new ArrayList<String>();
        Enumeration<String> en = textElements.keys();
        while (en.hasMoreElements()) {
            String key = en.nextElement();
            a.add(key);
        }
        Collections.sort(a);
        for (String key : a) {
            System.out.format("%s: %s\n", key, textElements.get(key));
        }
    }
    
    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        try {
            push(qName, atts);
            String path = buildPath();
            if (path.startsWith("TEI.text") && qName.equals("pb")) {
                if (sb.length() > 0) {
                    if (sb.charAt(sb.length()-1) != '\n') {
                        sb.append("\n");
                    }
                }
                String value = atts.getValue("facs");
                if (value == null) {
                    System.out.format("***%s:%d missing facs attribute on pb\n", filename, locator.getLineNumber());
                    sb.append("\nimage: Unknown\n");
                } else {
                    String facs = value.substring(1);
                    String url = folios.get(facs);
                    if (url == null) {
                        System.out.format("***%s:%d <pb facs=%s ...> does not map to a url\n", filename, locator.getLineNumber(), facs);
                        sb.append("\nimage: Unknown\n");
                    } else {
                        String imagefile = String.format("%s/%s", manuscriptDirectory, url); 
                        copyScript.append(String.format("cp %s .\n", imagefile));
                        htmlScript.append(String.format("<a target='_blank' href='file://%s'>%s</a><br/>\n", imagefile, url));
                        sb.append(String.format("\nimage: %s\n", imagefile));
                    }
                }
            } else if (path.equals("TEI.facsimile.surface.zone.graphic")) {
                String id = atts.getValue("xml:id");
                if (id == null) {
                    System.out.format("+++%s:%d missing xml:id attribute on zone.graphic\n", filename, locator.getLineNumber());
                } else {
                    String url = atts.getValue("url");
                    if (url == null) {
                        System.out.format("+++%s:%d missing url attribute on zone.graphic\n", filename, locator.getLineNumber());
                    } else {
                        folios.put(id, url);
                    }
                }
            } else if (path.startsWith("TEI.text")) {
                if (textElements.get(qName) == null) {
                    textElements.put(qName, String.format("%s:%d", shortName, locator.getLineNumber()));
                }
                
                String action = perm.get(qName);
                if (action != null) {
                    if (action.equals("allow")) {
                        permStack.add(Boolean.TRUE);
                    } else if (action.equals("break")) {
                        sb.append("\n");
                        String n = atts.getValue("n");
                        if (n == null) {
                            n = "Unknown";
                        }
                        sb.append(String.format("Line %s: ", n));
                        permStack.add(Boolean.FALSE);
                    } else {
                        sb.append(action);
                        permStack.add(Boolean.FALSE);
                    }
                } else {
                    permStack.add(Boolean.FALSE);
                }
            }
        } catch (Exception ex) {
            System.out.format("%d. %s\n", locator.getLineNumber(), ex.toString());
        }
    }
    
    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
        String path = buildPath();
        StringBuilder buffer = buffers.get(path);
        String value = null;
        if (buffer != null) {
            value = buffer.toString();
            buffer.setLength(0);
        } else {
            value = "";
        }
    
        if (value.trim().length() > 0) {
            if (path.contains("handNote")) {
                value = value.trim();
            }
            if (path.equals("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.objectDesc.supportDesc.condition")) {
                sb.append(String.format("condition: %s\n", value));
            } else if (path.equals("TEI.teiHeader.fileDesc.sourceDesc.msDesc.physDesc.handDesc.handNote.persName")) {
                sb.append(String.format("scribe name: %s\n", value));
            }
        }
        pop();
        
        if (permStack.size() > 0) {
            permStack.remove(permStack.size()-1);
        }
    }
    
    
    private boolean isPermitted () {
        boolean rc = true;
        
        if (permStack.size() > 0) {
            for (int i = permStack.size()-1; i >= 0; i--) {
                if (!permStack.get(i)) {
                    rc = false;
                    break;
                }
            }
        } else {
            rc = false;
        }
        
        return rc;
    }
    
    @Override
    public void characters (char[] ch, int start, int end) {
        String path = buildPath();
        StringBuilder buffer = buffers.get(path);
        if (buffer == null) {
            buffer = new StringBuilder();
            buffers.put(path, buffer);
        }
        String s = new String(ch, start, end);
        buffer.append(s);
        s = s.trim();
        if (s.length() > 0) {
            if (isPermitted()) {
                sb.append(s);
            }
        }
    }
    
    public static void convertToXML(String input, String tempOutput, String output) throws Exception {
        BufferedReader br = IOAssistance.openReader(input);
        BufferedWriter bw = IOAssistance.openWriter(tempOutput);
        
        String line;
        int lineno = 0;
        boolean closeManuscriptElement = false;
        boolean closeImageElement = false;
        bw.write("<manuscripts>\n");
        boolean conditionNext = false;
        while ((line=br.readLine()) != null) {
            line = line.trim();
            lineno++;
            if (line.length() == 0) {
                continue;
            }
            if (line.equals("penn0388.xml")) {
                line = "penn0388.xml";            
            }
            
            if (conditionNext) {
                conditionNext = false;
                if (!line.startsWith("condition")) {
                    bw.write(" <condition>Not specified</condition>\n");
                    bw.write("  <images>\n");
                }            
            }
            if (line.endsWith(".xml")) {
                conditionNext = true;
                if (closeManuscriptElement) {
                    bw.write("  </image>\n");
                    bw.write(" </images>\n");
                    bw.write("</manuscript>\n");
                    closeImageElement = false;
                } else {
                    closeManuscriptElement = true;
                }
                bw.write(String.format("<manuscript catalog=\"%s\">\n", line));
            } else if (line.startsWith("condition:")) {
                bw.write(String.format(" <condition>%s</condition>\n", line.substring(10).trim()));
                bw.write("  <images>\n");
            } else if (line.startsWith("image:")) {                
                if (closeImageElement) {
                    bw.write("  </image>\n");
                } else {
                    closeImageElement = true;
                }
                bw.write(String.format("  <image name=\"%s\">\n", line.substring(6).trim()));
            } else {            // hope is a line
                reLine.reset(line);
                if (reLine.find()) {                    
                    bw.write(String.format("   <line n=\"%s\"", reLine.getGroup(2)));
                    String flags = reLine.getGroup(1);
                    if (flags.equals("*")) {
                        bw.write(" corrected=\"yes\">");
                    } else if (flags.equals("x")) {
                        bw.write(" incomplete=\"yes\">");
                    } else if (flags.length() == 2) {
                        bw.write(" incomplete=\"true\" corrected=\"true\">");
                    } else {
                        bw.write(">");
                    }
                    bw.write(reLine.getGroup(3).trim());
                    bw.write("</line>\n");
                } else {
                    bw.write(String.format("   <comment>%s</comment>\n", line));
                }
            }
        }
        bw.write("  </image>\n");
        bw.write(" </images>\n");
        bw.write("</manuscript>\n");
        bw.write("</manuscripts>\n");
        br.close();
        bw.close();
        
        FixupOCRXML pop = new FixupOCRXML(tempOutput, output);
        pop.parse();
    }
}
