package qnd;

import org.sanskritlibrary.Regex;

public class SLPInfo {
    String sid;
    String preds;
    String root;
    String marker;
    String premarker;
    String slp;
    Regex re = new Regex("", "([aAiIuUfFxXeEoO])");
    Regex renasal = new Regex("", "([aAiIuUfFxXeEoO])(?![\\\\/^~])");

    public void computeSLP() {
        String s;
        String _premarker = "";
        String _root = "";
        String _marker = "";
        // put accents on the premarker
        //II.	Insert nasalization ... after premarker vowels other than those in the premarker sequences Yi, wu, or qu.
        if (premarker != null) {
            s = premarker;
            String prefix = "";
            if (premarker.startsWith("Yi") || premarker.startsWith("wu") || premarker.startsWith("qu")) {
                s = premarker.substring(2);
                prefix = premarker.substring(0, 2);
            }

            _premarker = applyAccent(s, '*', "~");
            _premarker = addNasalization(_premarker);
            _premarker = prefix + _premarker;
        }
        
        // put accents on the root.
        /*
         A.	The first character of the preds attribute of the entry element determines the accent of the root.
         1.	u --> /
         2.	a --> \
         3.	x --> do nothing
         4.	n --> do nothing [preds="none"]
         */
        if (root != null) {
            re.reset(root);
            int offset = 0;
            int cnt = 0;
            while (re.find()) {
                offset = re.getOffset(1);
                cnt++;
            }
            if (cnt > 1) {
                cnt = 0;
            }
            String prefix = root.substring(0, offset);
            String remainder = root.substring(offset);
            _root = prefix + applyAccent (remainder, preds.charAt(0), "");
        }

        if (marker != null) {
        // Insert nasalization after final vowel markers after their accents.
            _marker = applyAccent(marker, preds.charAt(2), "~");
            _marker = addNasalization(_marker);
        }
        
        slp = _premarker + _root + _marker;
        if (slp.length() == 0) {
            slp = "???";
        }
        
        System.out.format("%s(%s): %s + %s + %s --> %s\n", sid, preds, fixup(premarker), fixup(root), fixup(marker), slp);
    }
    
    private String fixup(String s) {
        if (s == null) {
            return "_";
        } else {
            return s;
        }
    }
    
    
    private String addNasalization (String s) {
        renasal.reset(s);
        int last = 0;
        StringBuilder sb = new StringBuilder();
        while (renasal.find()) {
            s = renasal.replace("$1~");
        }
        return s;
    }
    
    private String applyAccent(String s, char ch, String nasal) {
        StringBuilder sb = new StringBuilder();
        re.reset(s);
        String accent;
        int cnt = 0;
        // get the last vowel.
        String vowel = null;
        re.reset(s);
        while (re.find()) {
            accent = null;
            if (ch == 'u') {
                accent = "/";
            } else if (ch == 'a') {
                accent = "\\\\";
            } else if (ch == 's') {
                accent = "^";
            } else if (ch == '*') {
                accent = "";
            }
            int last = 0;
            if (accent != null) {
                //int offset = re.getOffset(1);
                //sb.append(s.substring(last, offset));
                //sb.append(re.replaceFirst("$1" + accent + nasal));
                //last = offset + 1;
                s = re.replace("$1" + accent + nasal);
            }
        }
        
        return s;
    }
}
