package qnd;

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Hashtable;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.SAXParser;
import org.sanskritlibrary.SAXParserAdapter;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

public class TranscoderPairs extends SAXParserAdapter {
    private Hashtable<String, Integer> pathsHashed = new Hashtable<String, Integer>();
    private Hashtable<String, ArrayList<String>> attvalues = new Hashtable<String, ArrayList<String>>();
    private ArrayList<ElementInfo> pathsSorted = new ArrayList<ElementInfo>();
    public TranscoderPairs (String filename)  throws Exception {
        SAXParserFactory factory = SAXParserFactory.newInstance();

        // Parse the input
        SAXParser saxParser = factory.newSAXParser();
        // See http://www.saxproject.org/apidoc/org/xml/sax/package-summary.html#package_description
        // Uncomment the following to turn on comment processing.
        // http://docs.oracle.com/javaee/1.4/tutorial/doc/JAXPSAX11.html
        // saxParser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
        try {
            saxParser.parse( new File(filename), this);
        } catch (Exception ex) {
            throw ex;
        }        
    }
    
    public ArrayList<ElementInfo> getPaths () {
        return pathsSorted;
    }
    
    public ArrayList<String> getAttValues (String att) {
        return attvalues.get(att);
    }
    
    @Override
    public void endDocument () {
        Enumeration<String> en = pathsHashed.keys();
        
        while (en.hasMoreElements()) {
            String key = en.nextElement();
            pathsSorted.add(new ElementInfo(key, pathsHashed.get(key)));
        }
        
        Collections.sort(pathsSorted, new ElementInfo("", 0));
        
        en = attvalues.keys();
        while (en.hasMoreElements()) {
            String key = en.nextElement();
            ArrayList<String> a = attvalues.get(key);
            Collections.sort(a);
        }
    }
    
    @Override
    public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
        push(qName, atts);
        String path = buildPath();
        if (pathsHashed.get(path) == null) {
            pathsHashed.put(path, locator.getLineNumber());
        }
        int len = atts.getLength();
        for (int i=0; i<len; i++) {
            String attName = atts.getQName(i);
            String key = String.format("%s@%s", path, attName);
            if (pathsHashed.get(key) == null) {
                pathsHashed.put(key, locator.getLineNumber());
                attvalues.put(key, new ArrayList<String>());
            }
            ArrayList<String> a = attvalues.get(key);
            if (a.size() < 20) {
                String value = atts.getValue(i);
                if (!a.contains(value)) {
                    a.add(value);
                }
            }
        }
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException {
        pop();
    }
    
    public class ElementInfo implements Comparator {
        public String qName;
        public int lineNumber;        
        
        public ElementInfo (String qName, int lineNumber) {
            this.qName = qName;
            this.lineNumber = lineNumber;
        }

        @Override
        public int compare(Object t, Object t1) {
            ElementInfo ei = (ElementInfo)t;
            ElementInfo ei1 = (ElementInfo)t1;
            
            return ei.qName.compareTo(ei1.qName);
        }
    }

}
