package qnd;

import java.io.BufferedWriter;
import java.io.File;
import java.util.Hashtable;
import org.sanskritlibrary.IOAssistance;
import org.sanskritlibrary.Regex;

public class ZoneUpdate {
    String correctionsDirectory;
    String zonesDirectory;
    Regex re = new Regex("", "url='([^']+)'");
    
    ZoneUpdate (String correctionsDirectory, String zonesDirectory) {
        if (!correctionsDirectory.endsWith("/")) {
            correctionsDirectory += "/";
        }
        if (!(zonesDirectory.endsWith("/"))) {
            zonesDirectory += "/";
        }
        this.correctionsDirectory = correctionsDirectory;
        this.zonesDirectory = zonesDirectory;
    }
    
    public void update () throws Exception {
        String plan = IOAssistance.readEntireFile(correctionsDirectory+"xml.txt");
        String[] planlines = plan.split("\n");
        int plancount = planlines.length;
        String currentZone = null;
        String currentZoneFile = null;
        Hashtable<String, String> h = new Hashtable<String, String>();
        boolean readZoneXML = true;
        for (int i=0; i<plancount; i++) {
            String pl = planlines[i].trim();
            if (readZoneXML) {
                if (pl.charAt(0) != '>') {
                    System.out.format("Expecting >... but saw %s\n", pl);
                    break;
                }
            
                currentZoneFile = zonesDirectory + pl.substring(1);
                File f = new File(currentZoneFile);
                if (!f.exists()) {
                    System.out.format("%s does not exist.\n", currentZoneFile);
                    break;
                }
                
                currentZone = IOAssistance.readEntireFile(currentZoneFile);
                
                // Make a backup
                BufferedWriter bw = IOAssistance.openWriter(currentZoneFile+".bak");
                bw.write(currentZone);
                bw.close();
                readZoneXML = false;
            } else {
                if (pl.length() == 0) {
                    readZoneXML = true;
                    updateZone(currentZoneFile, currentZone, h);
                } else {
                    String fn = correctionsDirectory + pl;
                    File f = new File(fn);
                    if (!f.exists()) {
                        System.out.format("Correction %s does not exist", fn);
                        break;
                    }
                    String s = IOAssistance.readEntireFile(fn);
                    String[] lines = s.split("\n");
                    if (!lines[0].contains("<surface>")) {
                        System.out.format("In %s. Expecting %s at line 1 but saw %s", fn, lines[0]);
                        break;
                    }
                    re.reset(lines[1]);
                    if (re.find()) {
                        h.put(re.getGroup(1), s);
                    } else {
                        System.out.format("In %s. Line 2 does not contain url: %s\n", fn, lines[1]);
                        return;
                    }
                }
            }                       
        }
    }
    
    private void updateZone (String currentZoneFile, String currentZone, Hashtable<String, String> h) throws Exception {
        BufferedWriter bw = IOAssistance.openWriter(currentZoneFile);
        String[] lines = currentZone.split("\n");
        int len = lines.length;
        for (int i=0; i<len; i++) {
          String line = lines[i];
          if (line.contains("<surface>")) {
              re.reset(lines[i+1]);
              if (re.find()) {                  // if this element need replacing.
                  String replacement = h.get(re.getGroup(1));
                  if (replacement != null) {
                      bw.write("<!-- Following surface was recropped. -->\n");
                      bw.write(replacement);
                      if (!replacement.endsWith("\n")) {
                          bw.write("\n");
                      }
                      h.remove(re.getGroup(1));
                      do {          // skip the replaced lines.
                        i++;
                        line = lines[i];
                      } while (!line.contains("</surface>"));
                      continue;
                  } 
              }
          }
          
          bw.write(line);
          bw.write("\n");          
        }
        bw.close();
        if (h.size() != 0) {
            System.out.format("%s: not all corrections found\n", currentZoneFile);
        }
        h.clear();
    }
}
